import regex as re
from lxml import etree
import glob
import os
from difflib import HtmlDiff
from IPython.display import display, Markdown
import xml.dom.minidom


def prettify(xml_string, indentation="  "):
    """Return a pretty-printed XML string"""
    return xml.dom.minidom.parseString(xml_string).toprettyxml(indent=indentation)


def remove_whitespace(text):
    # we need to remove the whitespace that comes with the indentation of pretty-printed xml
    text = re.sub(r'\n *', ' ', text)

    # reduce double spaces to one
    while re.search(r'  ', text):
        text = re.sub(r'  ', ' ', text)

    # escape character sequences which would be corrupted by whitespace removal rules
    text = re.sub(r'\. \. \.', '[!spaced_elipsis!]', text)

    # fix issues with whitespace before and after punctuation
    text = re.sub(r' ([.;,!?%:])( |$)', r'\1 ', text)
    # opening and closing punctutation, such as brackets
    text = re.sub(r'(\p{Ps}) ', r'\1', text)
    text = re.sub(r' (\p{Pe})', r'\1', text)
    # opening and closing quotes
    text = re.sub(r'(\p{Pi}) ', r'\1', text)
    text = re.sub(r' (\p{Pf})', r'\1', text)
    # slash
    text = re.sub(r' ?/ ?', r'/', text)

    # restore sequences
    text = text.replace('[!spaced_elipsis!]', r'. . .')
    return text.strip()

def fix_serialization_issues(text):
    # escape character sequences
    text = re.sub(r'\.\.\.', '[!elipsis!]', text)
    text = re.sub(r'\. \. \.', '[!spaced_elipsis!]', text)

    # remove duplicated punctuation
    text = re.sub(r'(\. ?){2}', r'.', text)

    # restore character sequences
    text = text.replace( '[!elipsis!]', '...')
    text = text.replace('[!spaced_elipsis!]', r'. . .')
    return text.strip()

def tei_to_input(tei_xml_doc):
    """
    Extract the original footnote strings from the <note> elements in a given TEI document and return a list of strings
    """
    root = etree.fromstring(tei_xml_doc)
    ref_list = []
    ns = {"tei": "http://www.tei-c.org/ns/1.0"}
    # iterate over the <note type="footnote"> and <listBibl> elements
    for xpath in ['*//tei:listBibl[@type="inText"]',
                  '*//tei:note[@type="footnote"]',
                  '*//tei:listBibl[@type="bibliography"]']:
        for element in root.findall(xpath, ns):
            localname = etree.QName(element).localname
            type = element.attrib['type'] if 'type' in element.attrib else None
            ref_parts = []
            if 'n' in element.attrib:
                ref_parts.append(element.attrib['n'])
            # iterate over the <bibl> elements
            for bibl in element.findall('tei:bibl', ns):
                # extract the text without xml tags, still contains all (collapsed) whitespace
                text = etree.tostring(bibl, method="text", encoding='utf-8').decode()
                text = remove_whitespace(text)
                ref_parts.append(text)
            if localname == "listBibl" and type != "inText":
                ref_list += ref_parts
            elif localname == "note" or (localname == "listBibl" and type == "inText"):
                ref_list.append(remove_whitespace(" ".join(ref_parts)))
    return ref_list


def compare_input_strings(tei_dir_path, ref_dir_path, html_output_path, target_url, refs_output_path=None):
    os.makedirs(html_output_path, exist_ok=True)
    for tei_bibl_path in glob.glob(f'{tei_dir_path}/*.xml'):
        base_name = os.path.basename(tei_bibl_path)
        id = os.path.splitext(base_name)[0]
        anystyle_input_path = f'{ref_dir_path}/{id}.txt'

        # load original raw reference strings
        with open(anystyle_input_path, 'r', encoding='utf-8') as f:
            anystyle_input_data = f.read().splitlines()

        # load tei bibl to check
        with open(tei_bibl_path, 'r', encoding='utf-8') as r:
            tei_xml_doc = r.read()

        # convert it to raw reference text
        tei_input_data = tei_to_input(tei_xml_doc)

        # postprocess to fix issues that should be fixed in the source or in the converter code
        tei_input_data = fix_serialization_issues(tei_input_data)

        # if an output path has been given, store the reconstructed raw references
        if refs_output_path:
            with open(f'{refs_output_path}/{id}.txt', 'w', encoding='utf-8') as w:
                w.write("\n".join(tei_input_data))

        # create files showing the diff between the reconstructed and the original raw strings
        html_diff = HtmlDiff().make_file(fromlines=anystyle_input_data,
                                         tolines=tei_input_data,
                                         fromdesc=anystyle_input_path,
                                         todesc=tei_bibl_path)

        # store the visual diff
        with open(f"{html_output_path}/{id}.diff.html", "w", encoding="utf-8") as f:
            f.write(html_diff)
            display(Markdown(f'Extracted and compared input data for {id}  ([See diff]({target_url}/{id}.diff.html))'))

# %%
